@@ -216,12 +216,11 @@ typedef struct Stmt
 {
   struct Stmt *next;
   Token *tokens;
+  Token *tokens_end;
   symbol *sym;
   unsigned char vis;
-  unsigned len : 12;
 } Stmt;
 
-#define alloc_comment(S,E) alloc_stmt (V_comment, S, E, 0)
 #define append_stmt(V, S) ((S)->next = *(V), *(V) = (S))
 
 static Stmt *decls;
@@ -325,7 +324,7 @@ tokenize (const char *ptr)
 		ptr++;
 
 	      block_comment:
-		eol = in_comment;
+		eol |= in_comment;
 		in_comment = 1;
 		for (; *ptr; ptr++)
 		  {
@@ -483,7 +482,7 @@ write_token (FILE *out, Token const *tok)
 static std::list<void *> heaps;
 
 static Stmt *
-alloc_stmt (unsigned vis, Token *tokens, Token *end, symbol *sym)
+alloc_stmt (unsigned vis, Token *tokens, Token *tokens_end, symbol *sym)
 {
   static unsigned alloc = 0;
   static Stmt *heap = 0;
@@ -502,7 +501,7 @@ alloc_stmt (unsigned vis, Token *tokens, Token *end, symbol *sym)
   stmt->next = 0;
   stmt->vis = vis;
   stmt->tokens = tokens;
-  stmt->len = end - tokens;
+  stmt->tokens_end = tokens_end;
   stmt->sym = sym;
 
   return stmt;
@@ -528,11 +527,11 @@ rev_stmts (Stmt *stmt)
 static void
 write_stmt (FILE *out, const Stmt *stmt)
 {
-  for (int i = 0; i < stmt->len; i++)
+  for (Token *tok = stmt->tokens; tok != stmt->tokens_end; ++tok)
     {
       if ((stmt->vis & V_mask) == V_comment)
 	fprintf (out, "//");
-      write_token (out, stmt->tokens + i);
+      write_token (out, tok);
       if ((stmt->vis & V_mask) == V_pred)
 	fputc (' ', out);
     }
@@ -551,7 +550,7 @@ write_stmts (FILE *out, const Stmt *stmts)
 /* Parse a line until the end, regardless of semicolons.  */
 
 static Token *
-parse_line_nosemi (Token *tok, Stmt **append_where)
+parse_line_nosemi (unsigned vis, Token *tok, Stmt **append_where)
 {
   Token *start = tok;
 
@@ -559,7 +558,7 @@ parse_line_nosemi (Token *tok, Stmt **append_where)
     tok++;
   while (!tok[-1].end);
 
-  Stmt *stmt = alloc_stmt (V_dot, start, tok, 0);
+  Stmt *stmt = alloc_stmt (vis, start, tok, 0);
   append_stmt (append_where, stmt);
 
   return tok;
@@ -580,12 +579,12 @@ parse_insn (Token *tok)
 	{
 	  if (is_keyword (tok, "file"))
 	    {
-	      tok = parse_line_nosemi (tok, &decls);
+	      tok = parse_line_nosemi (V_dot, tok, &decls);
 	      continue;
 	    }
 	  if (is_keyword (tok, "loc"))
 	    {
-	      tok = parse_line_nosemi (tok, &fns);
+	      tok = parse_line_nosemi (V_dot, tok, &fns);
 	      continue;
 	    }
 	}
@@ -594,7 +593,7 @@ parse_insn (Token *tok)
 	case K_comment:
 	  while (tok->kind == K_comment)
 	    tok++;
-	  stmt = alloc_comment (start, tok);
+	  stmt = alloc_stmt (V_comment, start, tok, 0);
 	  append_stmt (&fns, stmt);
 	  continue;
 
@@ -635,7 +634,7 @@ parse_insn (Token *tok)
       if (!tok[-1].end && tok[0].kind == K_comment)
 	{
 	  stmt->vis |= V_no_eol;
-	  stmt = alloc_comment (tok, tok + 1);
+	  stmt = alloc_stmt (V_comment, tok, tok + 1, 0);
 	  append_stmt (&fns, stmt);
 	  tok++;
 	}
@@ -658,7 +657,7 @@ parse_init (htab_t symbol_table, Token *tok, symbol *sym)
 	{
 	  while (tok->kind == K_comment)
 	    tok++;
-	  stmt = alloc_comment (start, tok);
+	  stmt = alloc_stmt (V_comment, start, tok, 0);
 	  append_stmt (&sym->stmts, stmt);
 	  start = tok;
 	}
@@ -681,7 +680,7 @@ parse_init (htab_t symbol_table, Token *tok, symbol *sym)
       if (!tok[-1].end && tok->kind == K_comment)
 	{
 	  stmt->vis |= V_no_eol;
-	  stmt = alloc_comment (tok, tok + 1);
+	  stmt = alloc_stmt (V_comment, tok, tok + 1, 0);
 	  append_stmt (&sym->stmts, stmt);
 	  tok++;
 	}
@@ -695,7 +694,7 @@ static Token *
 parse_file (htab_t symbol_table, Token *tok)
 {
   Stmt *comment = 0;
-  bool is_decl = false;
+  bool is_map_directive = false;
 
   if (tok->kind == K_comment)
     {
@@ -703,18 +702,33 @@ parse_file (htab_t symbol_table, Token *tok)
 
       while (tok->kind == K_comment)
 	{
+	  /* Instead of 'K_comment', a point could be made to have these be
+	     represented as their own 'Kind'.  */
 	  if (strncmp (tok->ptr, ":VAR_MAP ", 9) == 0
-	      || strncmp (tok->ptr, ":FUNC_MAP ", 10) == 0)
-	    /* GCC 'mkoffload's require these to be emitted in order of
-	       appearance; handle via 'decls'.  */
-	    is_decl = true;
+	      || strncmp (tok->ptr, ":FUNC_MAP ", 10) == 0
+	      || strncmp (tok->ptr, ":IND_FUNC_MAP ", 14) == 0)
+	    {
+	      is_map_directive = true;
+	      break;
+	    }
 	  tok++;
 	}
-      comment = alloc_comment (start, tok);
-      comment->vis |= V_prefix_comment;
+      if (start != tok)
+	{
+	  comment = alloc_stmt (V_comment, start, tok, 0);
+	  comment->vis |= V_prefix_comment;
+	}
     }
 
-  if (tok->kind == K_dotted)
+  if (is_map_directive)
+    {
+      /* GCC 'mkoffload' requires these to be emitted in order of appearance;
+	 handle via 'decls'.  */
+      if (comment)
+	append_stmt (&decls, comment);
+      tok = parse_line_nosemi (V_comment, tok, &decls);
+    }
+  else if (tok->kind == K_dotted)
     {
       if (is_keyword (tok, "version")
 	  || is_keyword (tok, "target")
@@ -723,12 +737,13 @@ parse_file (htab_t symbol_table, Token *tok)
 	{
 	  if (comment)
 	    append_stmt (&decls, comment);
-	  tok = parse_line_nosemi (tok, &decls);
+	  tok = parse_line_nosemi (V_dot, tok, &decls);
 	}
       else
 	{
 	  unsigned vis = 0;
 	  symbol *def = 0;
+	  bool is_decl = false;
 	  Token *start, *def_token = 0;
 
 	  for (start = tok;
@@ -1249,26 +1264,34 @@ This program has absolutely no warranty.\n",
 	  target_arg = preamble_target_arg;
 
 	  if ((strcmp ("sm_30", target_arg) == 0)
-	      || (strcmp ("sm_32", target_arg) == 0))
+	      || (strcmp ("sm_32", target_arg) == 0)
+	      || (strcmp ("sm_35", target_arg) == 0)
+	      || (strcmp ("sm_37", target_arg) == 0))
 	    {
-	      /* Starting with CUDA 11.0, "Support for Kepler 'sm_30' and
-		 'sm_32' architecture based products is dropped", and these may
-		 no longer be specified in '--gpu-name' of 'ptxas':
+	      /* In CUDA 11.0, "Support for Kepler 'sm_30' and 'sm_32'
+		 architecture based products is dropped", and in CUDA 12.0,
+		 "Kepler architecture support is removed" (that is, sm_35,
+		 sm_37), and these may no longer be specified in '--gpu-name'
+		 of 'ptxas':
 
 		     ptxas fatal   : Value 'sm_30' is not defined for option 'gpu-name'
 
 		     ptxas fatal   : Value 'sm_32' is not defined for option 'gpu-name'
 
+		     ptxas fatal   : Value 'sm_35' is not defined for option 'gpu-name'
+
+		     ptxas fatal   : Value 'sm_37' is not defined for option 'gpu-name'
+
 		 ..., but we need to continue supporting GCC emitting
 		 '.target sm_30' code, for example.
 
 		 Detecting the CUDA/'ptxas' version and the supported
 		 '--gpu-name' options is clumsy, so in this case, just use
-		 'sm_35', which is the baseline supported by all current CUDA
+		 'sm_50', which is the baseline supported by all current CUDA
 		 versions down to CUDA 6.5, at least.  */
 	      if (verbose)
 		fprintf (stderr, "Verifying %s code", target_arg);
-	      target_arg = "sm_35";
+	      target_arg = "sm_50";
 	      if (verbose)
 		fprintf (stderr, " with %s code generation.\n", target_arg);
 	    }
